import os
import random
import numpy as np

import torch
import pytorch_lightning as pl

from .utils import exceptions
from .llm import LLM

class AutoSeqML:
    def __init__(self, path):
        self._path = path
        
    def __str__(self):
        return f'training data path: {self._path}'
    
    def is_path_available(self):
        return os.path.exists(self._path)
    
    def items(self):
        return os.listdir(self._path)
    
    def seed(self, seed):
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        pl.seed_everything(seed)
    
    def downstream(self):
        llm = LLM(self._path)
        
    
if __name__=='__main__':
    print('Auto sequential ML package loaded!')


import os
from .weather import Weather

class PatternMining:
    def __init__(self, window_size=1):
        self.ws = window_size
    
    def split(self):
        pass
    
    def embed(self, content):
        lexicon = {}
        
    
    def encode(self, code:list):
        path = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(path, 'wordlist.txt'), 'r') as f:
            wl = [w.strip() for w in f.readlines()]
        
        return ' '.join([wl[c] for c in code])
    
if __name__=='__main__':
    print('test')


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from .eda import EDA

class CSV(EDA):
    def __init__(self, file, graph_path=''):
        super().__init__(file, graph_path)
        


import pandas as pd

from ..utils import check_ext

class Weather:
    def __init__(self, file):
        if not check_ext(file, 'weather'): return 'Invalid extensions or wrong data type'
        else:
            pass
    
    


import os
import random
# import wandb
import numpy as np
import pandas as pd

import torch
from torch.utils.data import DataLoader
from torchinfo import summary

import pytorch_lightning as pl
from pytorch_lightning.loggers import WandbLogger, TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping

from .dataset import Data
from .classify import BertClassifier as bc

class LLM:
    """
    Large language model downstream tasks.
    """
    def __init__(self, data_dir='', mode='classify', seed=-1, pretrained='bert-base-uncased'):
        """
        :param data_dir: string, path for training files
        :param mode: string, downstream mode, must be one of 'classify' or 'forecast'
        :param seed: integer, random seed, -1 for totally random
        :param pretrained: string, pretrained model for downstream task
        """
        assert data_dir, 'please provide file path for training.'
        
        self.param = {}
        self.param['path']          = data_dir
        self.param['mode']          = mode
        self.param['pretrained']    = pretrained
        
        self.param['input_size']    = 768 if mode=='classify' else 1536
        self.param['hidden_size']   = 0
        self.param['output_size']   = 2
        self.param['hidden_layers'] = 0
        self.param['learning_rate'] = 1e-5
        self.param['dropout']       = 0.5
        self.param['batch_size']    = 16

        if seed>-1:
            torch.manual_seed(seed)
            np.random.seed(seed)
            random.seed(seed)
            pl.seed_everything(seed)
            
    def __setitem__(self, key, value):
        self.param[key] = value
        
    def __getitem__(self, key):
        return self.param.get(key, 'param not found')

    def split_data(self, val_split=0.8, test_split=0.9):
        data = [file for file in os.listdir(self.param['path']) if file[-3:]=='csv']
        if len(data)==1:
            df = pd.read_csv(os.path.join(self.param['path'],data[0]))
            train, val, test = np.split(df, [int(val_split*len(df)), int(test_split*len(df))])
        elif len(data)==2:
            df = pd.read_csv(os.path.join(self.param['path'],list(filter(lambda x:'train' in x, data))[0]))
            train, val = np.split(df, [int(val_split*len(df))])
            test = pd.read_csv(os.path.join(self.param['path'],list(filter(lambda x:'test' in x, data))[0]))
        elif len(data)==3:
            train = pd.read_csv(os.path.join(self.param['path'],list(filter(lambda x:'train' in x, data))[0]))
            val = pd.read_csv(os.path.join(self.param['path'],list(filter(lambda x:'val' in x, data))[0]))
            test = pd.read_csv(os.path.join(self.param['path'],list(filter(lambda x:'test' in x, data))[0]))
        else:
            return 'too many files for training'
        
        return train, val, test
        
    def train(self, train, val, prefix=''):
        if self.param['mode']=='classify':
#             wandb.init(project='test')
#             wandb.config.batch_size = self.param['batch_size']
            
            if prefix: self.prefix = prefix
            else: self.prefix = 'test'

            train = Data(train, pretrained=self.param['pretrained'])
            self.param['output_size'] = train.labels

            train = DataLoader(train, batch_size=self.param['batch_size'], shuffle=True, num_workers=4)
            val = DataLoader(
                Data(val, pretrained=self.param['pretrained']),
                batch_size=self.param['batch_size'], shuffle=False, num_workers=4
            )

            checkpoint_callback_val_loss = ModelCheckpoint(
                save_top_k=1,
                monitor="val_loss"
            )
            early_stop_callback = EarlyStopping(
                monitor='val_loss',  # Metric to monitor for improvement
                min_delta=0.0,       # Minimum change in the monitored metric to be considered as an improvement
                patience=5,          # Number of epochs with no improvement after which training will be stopped
                verbose=False,
                mode='min'           # Whether the monitored metric should be minimized or maximized
            )

            trainer = pl.Trainer(
                devices=1,
                accelerator='gpu',
                max_epochs=-1,
                precision=16,
                log_every_n_steps=10,
                check_val_every_n_epoch=1,
                callbacks=[checkpoint_callback_val_loss, early_stop_callback],
                logger=TensorBoardLogger("lightning_logs", name=self.prefix)
#                 logger=WandbLogger(project=self.prefix, log_model=True)
            )

            model = bc(
                input_size    = self.param['input_size'],
                hidden_size   = self.param['hidden_size'],
                output_size   = self.param['output_size'],
                hidden_layers = self.param['hidden_layers'],
                learning_rate = self.param['learning_rate'],
                dropout       = self.param['dropout'],
                pretrained    = self.param['pretrained']
            )

#             print(summary(model, [(1,512), (1,512)], dtypes=[torch.cuda.IntTensor, torch.cuda.IntTensor]))

            trainer.fit(model, train, val)
#             wandb.finish()
        
        elif self.param['mode']=='forecast':
            pass
    
    def test(self, test):
        test = DataLoader(Data(test, pretrained=self.param['pretrained']), batch_size=self.param['batch_size'], shuffle=False, num_workers=0)
        pass
    
    def save(self):
        pass
        

if __name__=='__main__':
    llm = LLM('../data')
    llm.split_data()




class Encode:
    def __init__(self, coding):
        self.coding = coding
        with open('wordlist.txt', 'r') as f:
            self.wl = [w.strip() for w in f.readlines()]
            
    def encode(self):
        return ' '.join([self.wl[_] for _ in self.coding])
    
if __name__=='__main__':
    coding = [1,7,2,0,0,3,7,1,45,24,6,2]
    
    print(coding, Encode(coding).encode())


import os
from configobj import ConfigObj

PATH = os.path.dirname(os.path.abspath(__file__))
CONFIG = os.path.join(PATH, 'config.ini')

def check_ext(file, data_type=''):
    cfg = ConfigObj(CONFIG, encoding='utf8')
    ext = file.split('.')[-1]
    try:
        return ext in cfg['extensions'][data_type]
    except:
        exts = []
        for v in cfg['extensions'].values():
            exts.extend(list(v))
        return ext in exts


import torch
from transformers import BertTokenizer
from torch.utils.data import Dataset

class Data(Dataset):
    def __init__(self, df, ds_model='bert', pretrained='bert-base-uncased'):
        if ds_model=='bert':
            tokenizer = BertTokenizer.from_pretrained(pretrained)
            self.x = [
                tokenizer(
                    code,
                    padding='max_length',
                    max_length=512,
                    truncation=True,
                    return_tensors='pt'
                ) for code in df['encode']
            ]
        elif ds_model=='gpt3':
            pass
        elif ds_model=='llama':
            pass
        
        label_dict = {k:v for v,k in enumerate(sorted(df['label'].unique().tolist()))}
        y = [label_dict[l] for l in df['label'].tolist()]
        self.y = torch.tensor(y)
        self.labels = len(label_dict)
    
    def __len__(self):
        return len(self.x)
    
    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]
    
if __name__=='__main__':
    import pandas as pd
    df = pd.read_csv('../../data/train.csv')
    print(Data(df))



from .csv import CSV
from .txt import TXT
from ..utils import check_ext

def EDA(file, graph_path=''):
    if not check_ext(file): return 'file extension unacceptable'

    ext = file.split('.')[-1]
    if ext=='csv':
        return CSV(file,graph_path)
    elif ext=='txt':
        return TXT(file,graph_path)



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

class EDA:
    def __init__(self, file, graph_path=''):
        self.path = os.path.dirname(os.path.abspath(file))
        self.graph_path = graph_path if graph_path else self.path
        if not os.path.exists(self.graph_path): raise FileNotFoundError
        self.df = pd.read_csv(file)
        self.columns = self.df.select_dtypes(include=float).columns
        
    def report(self):
        return self.df.describe(percentiles=[0.1,0.25,0.5,0.75,0.9]).transpose()
    
    def show(self, col=''):
        if not col: raise ValueError('Select a column to show 1d figure')
        fig = self.df[col].plot(figsize=(5,3),title=col).get_figure()
        fig.savefig(os.path.join(self.graph_path,'1d.png'))
           
    def show2d(self, col1='', col2=''):
        if not col1 or not col2: raise ValueError('Need 2 columns to form 2d figure')
        df1 = self.df[col1]
        df2 = self.df[col2]
        vmax = int(max(max(df1),max(df2))//50+1)*50
        plt.hist2d(df1, df2, bins=100, vmax=vmax, cmap='gray')
        plt.colorbar()
        plt.savefig(os.path.join(self.graph_path,'2d.png'))

    def save_graph(self):
        cols = self.columns
        for i,col in enumerate(cols):
            fig = self.df[col].plot(figsize=(5,3),title=col).get_figure()
            fig.savefig(os.path.join(self.graph_path,f'{col}.png'))
            plt.close(fig)
    
    def set_range(self, col, mini=np.nan, maxi=np.nan):
        if mini is np.nan or maxi is np.nan:
            return 'please give a exact range to set the range of data'
        for idx, value in enumerate(self.df[col]):
            if value > maxi:
                self[idx,col] = maxi
                continue
            if value < mini:
                self[idx,col] = mini
                continue
    
    def __getitem__(self, index):
        return self.df.at[index]
    
    def __setitem__(self, index, value):
        self.df.at[index] = value


import pathlib
from setuptools import setup, find_packages
 
HERE = pathlib.Path(__file__).parent.resolve()
 
PACKAGE_NAME = "autoseqml"
AUTHOR = "Kun-Lin, Ho"
AUTHOR_EMAIL = "m11015064@mail.ntust.edu.tw"
URL = "https://github.com/kurokumasan/AutoSeqML"
DOWNLOAD_URL = ""#https://pypi.org/project/Find-sitemap/"
 
LICENSE = "MIT"
VERSION = "0.1.1"
DESCRIPTION = "A tool for dealing sequential data for machine learning"
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding="utf8")
LONG_DESC_TYPE = "text/markdown"
 
requirements = (HERE / "requirements.txt").read_text(encoding="utf8")
INSTALL_REQUIRES = [s.strip() for s in requirements.split("\n")]
 
dev_requirements = (HERE / "dev_requirements.txt").read_text(encoding="utf8")
EXTRAS_REQUIRE = {"dev": [s.strip() for s in dev_requirements.split("\n")]}
 
CLASSIFIERS = [f"Programming Language :: Python :: 3.{str(v)}" for v in range(7, 12)]
PYTHON_REQUIRES = ">=3.7"
 
setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    license=LICENSE,
    author_email=AUTHOR_EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    packages=find_packages(),
    package_data={'': ['config.ini']},
    include_package_data=True,
    classifiers=CLASSIFIERS,
)



# pretrained model
from transformers import BertModel

# torch package
import torch
from torch import nn
from torch.optim import AdamW

# torch ligntning
import pytorch_lightning as pl

class BertClassifier(pl.LightningModule):
    
    def __init__(
        self,
        input_size=768,
        hidden_size=0,
        output_size=2,
        hidden_layers=0,
        learning_rate=1e-5,
        dropout=0.5,
        pretrained='bert-base-uncased'
    ):
        super().__init__()
        print(f'input size: {input_size}, hidden layers: {hidden_layers}, hidden size: {hidden_size}, output size: {output_size}')
        print(f'Using pretrained model: {pretrained}')
        self.bert = BertModel.from_pretrained(pretrained)
        self.dropout = nn.Dropout(dropout)
        self.linear = nn.Linear(input_size, output_size)
        self.fc_layers = nn.ModuleList()
        
        self.lr = learning_rate
        self.loss_fn = nn.CrossEntropyLoss()
        self.relu = nn.ReLU(inplace=True)
        
        if hidden_layers:
            for i in range(hidden_layers):
                in_size = hidden_size if i else input_size
                self.fc_layers.append(nn.Linear(in_size, hidden_size))
            self.fc_layers.append(nn.Linear(hidden_size, output_size))
        else:
            self.fc_layers.append(nn.Linear(input_size, output_size))
            
    def forward(self, ids, mask):
        _, x = self.bert(ids, mask, return_dict=False)
        x = self.dropout(x)
        for i,fc in enumerate(self.fc_layers):
#             if i==len(self.fc_layers)-1:
#                 x = fc(x)
#             else:
            x = fc(x)
            x = self.relu(x)
        return x
    
    def training_step(self, batch, batch_idx):
        ids, mask = batch[0]['input_ids'].squeeze(1), batch[0]['attention_mask']
        pred = self(ids, mask)
        loss = self.loss_fn(pred, batch[1])
        self.log('training loss', loss)
        
        acc = self.accuracy(pred, batch[1])
        self.log('train_acc', acc, on_epoch=True)
        return loss
    
    def validation_step(self, batch, batch_idx):
        ids, mask = batch[0]['input_ids'].squeeze(1), batch[0]['attention_mask']
        pred = self(ids, mask)
        loss = self.loss_fn(pred, batch[1])
        self.log('val_loss', loss)
        
        acc = self.accuracy(pred, batch[1])
        self.log('val_acc', acc, on_epoch=True)
    
    def test_step(self, batch, batch_idx):
        ids, mask = batch[0]['input_ids'].squeeze(1), batch[0]['attention_mask']
        pred = self(ids, mask)
        loss = self.loss_fn(pred, batch[1])
        self.log('test_loss', loss)
        
        acc = self.accuracy(pred, batch[1])
        self.log('test_acc', acc, on_epoch=True)
    
    def configure_optimizers(self):
        return AdamW(self.parameters(), lr=self.lr, weight_decay=0.00001)

    def accuracy(self, pred, label):
        correct = (pred.argmax(dim=1)==label).sum().item()
        return correct/len(pred)

if __name__=='__main__':
    print('test')
